/*
 * Provide support for creating and accessing pipes.
 *
 * Copyright (C) 2020 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/sys/types.h>

#include <stdio.h>
#include <stdlib.h>

#include <fsserver/paging.h>
#include <fsserver/pipe_resource.h>
#include <fsserver/resource_server.h>
#include "pipe_opener_server.h"



/* A resource providing pipe opening capabilities. */

class PipeOpenerResource : public Resource, public PipeOpener
{
protected:
  Paging *_paging;

public:
  explicit PipeOpenerResource(Paging *paging = 0)
  : Resource(), _paging(paging)
  {
  }

  /* Server details. */

  int expected_items()
  { return PipeOpener_expected_items; }

  ipc_server_handler_type handler()
  { return (ipc_server_handler_type) handle_PipeOpener; }

  void *interface()
  { return static_cast<PipeOpener *>(this); }

  /* Operation methods. */

  long pipe(l4_cap_idx_t *reader, l4_cap_idx_t *writer)
  {
    long err;

    /* Provide access to pipe regions. */

    PipePagingAccessor *accessor = new PipePagingAccessor(_paging);

    /* Obtain resources to main reading and writing state for the clients. */

    PipeResource *reader_resource = new PipeResource(accessor, false, 4096);
    PipeResource *writer_resource = new PipeResource(accessor, true, 4096);

    /* Employ servers for the resources. */

    ResourceServer reader_server(reader_resource);
    ResourceServer writer_server(writer_resource);

    err = reader_server.start_thread();
    if (err)
      return err;

    err = writer_server.start_thread();
    if (err)
      return err;

    *reader = reader_server.config()->server;
    *writer = writer_server.config()->server;
    printf("Returning pipe endpoints.\n");
    return L4_EOK;
  }
};



int main(void)
{
  /* Allocate pages for pipe data. */

  Paging *paging = new Paging(new Pages(0));

  /* Initialise and register a new server object. */

  PipeOpenerResource resource(paging);
  ResourceServer server(&resource);

  if (server.bind("export"))
  {
    printf("Could not bind thread.\n");
    return 1;
  }

  /* Enter the IPC server loop. */

  printf("Starting pipeserver...\n");
  server.start();
  return 0;
}
